/*-----------------------------------------------------------------------

File  : ccl_grounding.c

Author: Stephan Schulz

Contents
 
  Functions for implementing grounding of near-propositional clause
  sets. Note that all functons here require a fairly static term bank
  - don't do any freeing or rewriting intermingeled with grounding, it
  may well smoke and burn!

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Tue May 29 02:25:12 CEST 2001
    New

-----------------------------------------------------------------------*/

#include "ccl_grounding.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: varsetinstapply()
//
//   Given a complete VarSetInst data structure, instaniate the
//   variables apropriately.
//
// Global Variables: -
//
// Side Effects    : Changes instantiations of variables
//
/----------------------------------------------------------------------*/

void varsetinstapply(VarSetInst_p varinst)
{
   long i;
   
   for(i=0; i< varinst->size; i++)
   {
      varinst->cells[i].variable->binding =
	 PStackElementP(varinst->cells[i].alternatives,
			varinst->cells[i].position);      
   }
}

/*-----------------------------------------------------------------------
//
// Function: varsetinstclear()
//
//   Clear variables instatiated by a VarSetInst structure.
//
// Global Variables: -
//
// Side Effects    : Changes instantiations of variables
//
/----------------------------------------------------------------------*/

void varsetinstclear(VarSetInst_p varinst)
{
   long i;
   
   for(i=0; i< varinst->size; i++)
   {
      varinst->cells[i].variable->binding = NULL;
   } 
}


/*-----------------------------------------------------------------------
//
// Function: varsetinstinitialize()
//
//   Initialize a VarSetInst structure to represent the initial ground
//   substitution (PStacks with alternatives have to be present!).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void varsetinstinitialize(VarSetInst_p varinst)
{
   long i;
   
   for(i=0; i< varinst->size; i++)
   {
      assert(varinst->cells[i].alternatives);
      varinst->cells[i].position =
	 PStackGetSP(varinst->cells[i].alternatives)-1;
   } 
}


/*-----------------------------------------------------------------------
//
// Function: varsetinstnext()
//
//   Switch to the next substitution, return false if substitution
//   space is exhausted (true otherwise).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool varsetinstnext(VarSetInst_p varinst)
{
   long i;
   
   for(i=0; i< varinst->size; i++)
   {      
      if(varinst->cells[i].position)
      {
	 varinst->cells[i].position--;
	 return true;
      }
      varinst->cells[i].position =
	 PStackGetSP(varinst->cells[i].alternatives)-1;
   } 
   return false;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: EqnEqlitRecode()
//
//   Recode an equational literal as a non-equational one using
//   $eq(l,r)=T. Return true if recoded, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes the literal in place!
//
/----------------------------------------------------------------------*/

bool EqnEqlitRecode(Eqn_p lit)
{
   if(EqnIsEquLit(lit))
   {
      Term_p t = TermDefaultCellAlloc();
      
      t->f_code = SigGetEqnCode(lit->bank->sig, true);
      t->args = TermArgArrayAlloc(2);
      t->arity = 2;
      t->args[0] = lit->lterm;
      t->args[1] = lit->rterm;
      t = TBTermtopInsert(lit->bank, t);
      TermReleaseRef(&(lit->lterm));
      TermReleaseRef(&(lit->rterm));
      TermGetRef(&(lit->lterm), t);
      TermGetRef(&(lit->rterm), lit->bank->true_term);
      EqnDelProp(lit, EPIsEquLiteral);
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseEqlitRecode()
//
//   Recode a potential equational clause to a non-equational
//   one. Return true if conversion took place.
//
// Global Variables: -
//
// Side Effects    : Changes clause
//
/----------------------------------------------------------------------*/

bool ClauseEqlitRecode(Clause_p clause)
{
   Eqn_p handle;
   bool  res=false, tmp;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      tmp = EqnEqlitRecode(handle);
      res = res | tmp;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetEqlitRecode()
//
//   Recode all clauses in set, return number of conversions.
//
// Global Variables: -
//
// Side Effects    : Changes clauses
//
/----------------------------------------------------------------------*/

long ClauseSetEqlitRecode(ClauseSet_p set)
{
   long     res = 0;
   Clause_p handle;

   for(handle = set->anchor->succ; handle!= set->anchor; handle =
	  handle->succ)
   {
      if(ClauseEqlitRecode(handle))
      {
	 res++;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: VarSetInstAlloc() 
//
//   Create a VarSetInst for all variables occurring in clause. Does
//   not allocate the PStacks needed!
//
// Global Variables: -
//
// Side Effects    : Memory allocation
//
/----------------------------------------------------------------------*/

VarSetInst_p VarSetInstAlloc(Clause_p clause)
{
   PTree_p      tree = 0;
   VarSetInst_p handle = VarSetInstCellAlloc();
   long         i;
   Term_p       var;
   
   handle->size = ClauseCollectVariables(clause, &tree);
   if(handle->size)
   {
      handle->cells = SizeMalloc(handle->size*sizeof(VarInstCell));
      
      for(i=0; i<handle->size; i++)
      {
	 assert(tree);

	 var = PTreeExtractRootKey(&tree);
	 handle->cells[i].variable = var;
	 handle->cells[i].alternatives = NULL;
      }
      assert(!tree);
   }
   else
   {
      handle->cells = NULL;
   }
   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: VarSetInstFree()
//
//   Free a VarSetInst. Does not free the PStacks()!
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void VarSetInstFree(VarSetInst_p junk)
{
   assert(junk);

   if(junk->size)
   {
      assert(junk->cells);
      SizeFree(junk->cells, junk->size*sizeof(VarInstCell));      
   }
   else
   {
      assert(!junk->cells);
   }   
   VarSetInstCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: SigCountNonConstFSymbols()
//
//   Return the number of real non-constant function symbols in the
//   signature. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long SigCountNonConstFSymbols(Sig_p sig)
{
   FunCode i;
   long res=0;

   for(i=sig->internal_symbols+1; i<=sig->f_count; i++)
   {
      if(!SigIsAnyFuncPropSet(sig, i, FPPredSymbol|FPSpecial))
      {
	 if(SigFindArity(sig,i))
	 {
	    res++;
	 }
      }
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: SigCollectConstantTerms()
//
//   Push terms corresponding to all constants in sig onto the
//   stack. If sig contains no constant, insert a new skolem
//   constant. 
//
// Global Variables: -
//
// Side Effects    : Potentially changes sig and bank
//
/----------------------------------------------------------------------*/

long SigCollectConstantTerms(TB_p bank, PStack_p stack)
{
   FunCode i;
   long res = 0;
   Term_p tmp, found;
   
   for(i=bank->sig->internal_symbols+1; i<=bank->sig->f_count; i++)
   {
      if(!SigQueryFuncProp(bank->sig, i, FPSpecial) &&
	 SigFindArity(bank->sig,i)==0)
      {
	 tmp = TermConstCellAlloc(i);
	 found = TBTermtopInsert(bank, tmp);
	 PStackPushP(stack, found);
	 res++;
      }
   }
   if(!res)
   {
      OUTPRINT(1, "# No constant in specification, "
	       "added new Skolem constant\n");
      i = SigGetNewSkolemCode(bank->sig, 0);
      tmp = TermConstCellAlloc(i);
      found = TBTermtopInsert(bank, tmp);
      PStackPushP(stack, found);
      res++;
   }
   return res;
}




/*-----------------------------------------------------------------------
//
// Function: ClauseCreateGroundInstances()
//
//   Create all non-tautological ground instances of clause described
//   by inst. Return the term identifier of the largest literal
//   created.
//
// Global Variables: -
//
// Side Effects    : Memory operations, changes inst.
//
/----------------------------------------------------------------------*/

long ClauseCreateGroundInstances(TB_p bank, Clause_p clause,
				 VarSetInst_p inst, ClauseSet_p groundset)
{
   long ident = 0;
   bool next = true;
   Eqn_p handle;
   Clause_p new;

   varsetinstinitialize(inst);
   
   while(next)
   {
      varsetinstapply(inst);
      handle = EqnListCopy(clause->literals, bank);
      EqnListRemoveDuplicates(handle, TBTermEqual);
      if(EqnListIsTrivial(handle))
      {
	 EqnListFree(handle);
      }
      else
      {
	 new = ClauseAlloc(handle);
	 for(handle = new->literals; handle; handle = handle->next)
	 {
	    ident = MAX(ident, handle->lterm->entry_no);
	 }
	 ClauseSetInsert(groundset, new);
      }
      next = varsetinstnext(inst);
   }
   varsetinstclear(inst);

   return ident;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetCreateGroundInstances()
//
//   Create all ground instances of set and put them into
//   groundset. Return the term identifier of the largest literal
//   created. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseSetCreateGroundInstances(TB_p bank, ClauseSet_p set,
				    ClauseSet_p groundset)
{
   PStack_p     default_terms = PStackAlloc();
   Clause_p     handle;
   VarSetInst_p inst;
   long         i, res = 0, tmp;

   SigCollectConstantTerms(bank, default_terms);

   for(handle = set->anchor->succ; handle!= set->anchor; handle =
	  handle->succ)
   {
      inst = VarSetInstAlloc(handle);
      for(i=0; i<inst->size; i++)
      {
	 inst->cells[i].alternatives = default_terms;
      }      
      tmp = ClauseCreateGroundInstances(bank, handle, inst, groundset);
      res = MAX(res, tmp);
      VarSetInstFree(inst);
   }
   PStackFree(default_terms);

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintDimacs()
//
//   Print a clause in DIMACS format.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintDimacs(FILE* out, Clause_p clause)
{
   Eqn_p handle;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(EqnIsPositive(handle))
      {
	 fprintf(stdout, "  %ld", handle->lterm->entry_no);
      }
      else
      {
	 fprintf(stdout, " -%ld", handle->lterm->entry_no);
      }
   }
   fputs(" O\n", out);
}

/*-----------------------------------------------------------------------
//
// Function: ClauseSetPrintDimacs()
//
//   Print a clause set in DIMACS format
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClauseSetPrintDimacs(FILE* out, ClauseSet_p set, long
			  max_literal)
{
   Clause_p handle;

   fprintf(out, "p cnf %ld %ld\n", max_literal, set->members);
   
   for(handle = set->anchor->succ; handle != set->anchor; handle =
	  handle->succ)
   {
      ClausePrintDimacs(out, handle);
   }
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


